#include "static_algorithms.h"
#include "decorateur_tache.h"
#include "algorithm_ext.h"
#include "cedule_statique.h"
#include "Tache.h"
#include "predicats_statiques.h"
#include "type_list.h"
#include "outils_numeriques_statiques.h"
#include "ordonnancement_cycles.h"
#include <algorithm>
#include <iterator>
#include <iostream>
#include <cassert>
#include <vector>
#include <iosfwd>
#include <list>
using namespace std;

template <class T, T N>
   ostream& operator<<(ostream &os, const integral_constant<T, N> &n) {
      return os << n.value;
   }

class afficher {
   ostream &os;
   const char *delim;
public:
   afficher(ostream &os, const char *delim = "") noexcept : os{ os }, delim{ delim } {
   }
   template <class T>
      void operator()(const T &arg) const {
         os << arg << delim;
      }
   template <>
      void operator()(const type_list<> &) const {
      }
};

template<class TList>
   void ceduler(construct_taskset<TList> taskset_constructor) {
      enum { CYCLE_MINEUR = calculer_cycle_mineur<TList>::value };
      enum { CYCLE_MAJEUR = calculer_cycle_majeur<TList>::value };
#ifdef _DEBUG
      cout << "Cycle mineur: " << CYCLE_MINEUR << endl
           << "Cycle majeur: " << CYCLE_MAJEUR << endl;
#endif
      //
      // Le taux d'occupation du task set est-il infrieur  1? C'est une condition
      // ncessaire (mais pas suffisante) pour que le task set puisse tre cdul...
      //
      static_assert(
         evaluer_occupation_ordonnancement<CYCLE_MAJEUR, TList>::value <= CYCLE_MAJEUR,
         "Le task set ne peut etre ordonnance. Cause: occupation superieure a 1"
      );
      static_assert(
         ordonnancable<TList>::value,
         "Impossible de construire un ordonnancement periodique statique"
      );
#ifdef _DEBUG
      cout << "Occupation theorique du task set: "
           << evaluer_occupation_ordonnancement<calculer_cycle_majeur<TList>::value, TList>::taux_occupation() * 100
           << '%' << endl;
#endif
      //
      // Gnrer la liste des tches
      //
      auto lst = taskset_constructor.execute();
#ifdef _DEBUG
      cout << "Taches a couvrir (P, C), portrait dynamique :" << endl << '\t';
      copy(begin(lst), end(lst), ostream_iterator<Tache_>(cout, "\n\t"));
      cout << endl;
#endif
      cedule_statique<CYCLE_MINEUR, CYCLE_MAJEUR> ced;
      for (const auto &tache : lst)
         if (!ced(tache)) {
            cerr << "incapable d'ordonnancer " << tache << endl;
            break;
         }
      cout << ced << endl;
   }


template <class TList>
   void tester_ceduleur_statique() {
      cout << "\n-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n"
           << "Taches a ordonnancer (P, C), portrait statique :\n\t";
      static_for_each<TList>::appliquer(afficher_tache{ cout, "\n\t" });
      cout<< endl;
      //
      // Construire une cdule
      //
      ceduler(construct_taskset<TList>());
   }

//template <class TList>
//   void describe_schedule(ostream &os) {
//      using current_task = static_head_t<TList>;
//      using base_schedule = typename ordonnancable<TList>::empty_schedule;
//      using result_schedule = typename ordonnancable<TList>::type;
//      using ongoing_scheduling = static_schedule_one<current_task, base_schedule>;
//      //
//      //
//      //
//      os << "Base (empty) schedule, size " << length_<base_schedule>::value << " :\n\t";
//      static_for_each<base_schedule>::appliquer(afficher{ os, " " });
//      os << "\n\n";
//      os << "Ongoing schedule, size " << length_<ongoing_scheduling::type>::value << " :\n\t";
//      static_for_each<typename ongoing_scheduling::type>::appliquer(afficher{ os, " " });
//      os << "\n\n";
//      os << "Beginning, middle, end, periods remaining :\n\t";
//      os << length_<typename ongoing_scheduling::beginning_part>::value << ','
//         << length_<typename ongoing_scheduling::middle_part>::value << ','
//         << length_<typename ongoing_scheduling::end_part>::value << '('
//         << length_<typename ongoing_scheduling::remaining_periods>::value
//         << ')' << "\n\n";
//      //os << "Resulting schedule, size " << length_<result_schedule>::value << " :\n\t";
//      //static_for_each<result_schedule>::appliquer(afficher{ os, " " });
//      os << endl << endl;
//   }

int main() {
   //
   // L'ensemble de tches priodiques statiques que nous souhaitons ordonnancer
   //
   using task_set_11_1 = type_list<
      Tache<25, 10>, Tache<25, 8>, Tache<50, 5>, Tache<50, 4>, Tache<100, 2>
   >;
   using task_set_test = type_list<Tache<6, 4>, Tache<12, 3>>;
   using task_set_banal = type_list<Tache<6, 4>>;
   using task_set_limite = type_list<Tache<5, 5>>;
   using task_set_illegal = type_list<Tache<5, 6>>;
   //describe_schedule<task_set_11_1>(cout);
   tester_ceduleur_statique<task_set_11_1>();
   //tester_ceduleur_statique<task_set_test>(); // illgal mais pas vident
   //tester_ceduleur_statique<task_set_banal>();
   //tester_ceduleur_statique<task_set_limite>();
   //tester_ceduleur_statique<task_set_illegal>(); // immdiat  dtecter
}

