#ifndef ORDONNANCEMENT_CYCLES_H
#define ORDONNANCEMENT_CYCLES_H

#include "type_list.h"
#include "static_algorithms.h"
#include "Tache.h"

//
// calculer_cycle_majeur<TList>
// - calcule le cycle majeur des tches dans la liste TList
// - le fruit de calculer_cycle_majeur<TList> est calculer_cycle_majeur<TList>::value,
//   la valeur entire du cycle majeur de TList, exprim en nombre d'units discrtes
//   de temps
//
// Le cycle majeur d'un ensemble de tches priodiques dont les paramtres
// (P, C) sont connus a priori est le plus petit commun multiple (PPCM) des
// priodes (P) de ces tches.
//
// Le cycle majeur d'un ordonnancement statique de tches priodiques est
// le plus petit intervalle d'units de temps menant  un cycle complet
// d'excution pour toutes les tches impliques. Cela signifie qu'il n'y
// a pas lieu de prvoir un ordonnancement statique plus long que la longueur
// du cycle majeur pour l'ensemble des tches impliques: en pratique, ce
// cycle se rptera, tout simplement. On peut dire que le cycle majeur d'un
// ensemble de tches est la priode de l'ensemble examin de manire globale.
//
template <class TList>
   struct calculer_cycle_majeur {
   private:
      using periodes = static_transform_t<extract_period, TList>;
      using periode_tete = extract_period_t<static_head_t<TList>>;
   public:
      enum {
         value = static_accumulate_t<static_ppcm, periode_tete, periodes>::value
      };
   };

//
// calculer_cycle_mineur<TList>
// - calcule le cycle mineur des tches dans la liste TList
// - le fruit de calculer_cycle_mineur<TList> est calculer_cycle_mineur<TList>::value,
//   la valeur entire du cycle mineur de TList, exprim en nombre d'units discrtes
//   de temps
//
// Le cycle mineur d'un ensemble de tches priodiques dont les paramtres
// (P, C) sont connus a priori est le plus grand commun diviseur (PGCD) des
// priodes (P) de ces tches.
//
// Le cycle mineur d'un ordonnancement statique de tches priodiques est
// le plus petit intervalle d'units de temps au bout duquel une dcision doit
// ncessairement tre prise par l'ordonnanceur. videmment, le mot dcision
// est quelque peu abusif ici, du fait que l'ordonnancement est statique, mais
// l'ide est que le dbut d'un cycle mineur est un point sensible pour le
// lancement (Release) d'une tche priodique; connatre le cycle mineur permet
// de construire un ordonnancement de manire plus efficace, en rduisant les
// moments candidats  provoquer un lancement de tche.
//
// En pratique, une tche en excution ne peut chevaucher le moment o on
// passe d'un cycle mineur  un autre; les dbuts de cycle mineurs sont aussi
// des dbuts de tches
//
template <class TList>
   struct calculer_cycle_mineur
      : static_accumulate_t<
           static_pgcd, extract_period_t<static_head_t<TList>>, static_transform_t<extract_period, TList>
        >
   {
   };

#endif

