#include "static_algorithms.h"
#include <iostream>
using namespace std;

namespace {
   void tester_count_consecutive() {
      using tl_test = type_list<int, int, int, int, short>;
      static_assert(count_consecutive_beginning<short, tl_test>::value == 0, "count_consecutive<short, tl_test>::value == 0");
      static_assert(count_consecutive_beginning<int, tl_test>::value == 4, "count_consecutive<int, tl_test>::value == 4");
      static_assert(count_consecutive_beginning<float, tl_test>::value == 0, "count_consecutive<float, tl_test>::value == 0");
   }
   void tester_count_if() {
      using tl_test = type_list<int, short, int, int, int, short>;
      static_assert(static_count_if<is_<short>::type, tl_test>::value == 2, "static_count_if<is_<short>, tl_test>::value == 2");
      static_assert(static_count_if<is_<int>::type, tl_test>::value == 4, "static_count_if<is_<int>, tl_test>::value == 4");
      static_assert(static_count_if<is_<float>::type, tl_test>::value == 0, "static_count_if<is_<float>, tl_test>::value == 0");
   }
   void tester_indice_par_type() {
      using tl_test = type_list<int, short, int, int, char, short>;
      static_assert(indice_par_type<short, tl_test>::value == 1, "indice_par_type<tl_test, short>::value == 1");
      static_assert(indice_par_type<int, tl_test>::value == 0, "indice_par_type<tl_test, int>::value == 0");
      static_assert(indice_par_type<char, tl_test>::value == 4, "indice_par_type<tl_test, char>::value == 4");
      static_assert(indice_par_type<float, tl_test>::value == -1, "indice_par_type<tl_test, float>::value == -1");
   }
   void tester_inserer() {
      using tl_test = type_list<int, short, int, int, char, float>;
      using tl_test_post = static_inserer_fin<void, tl_test>::type;
      static_assert(!is_same<tl_test, tl_test_post>::value, "is_same<tl_test, tl_test_post>::value");
   }
   void tester_inserer_fin() {
      using tl_test = type_list<int, short, int, int, char, float>;
      static_assert(indice_par_type<float, tl_test>::value == 5, "avant insertion, indice_par_type<tl_test, float>::value == 5");
      using tl_test_post = static_inserer_fin<void, tl_test>::type;
      static_assert(indice_par_type<void, tl_test_post>::value == 6, "apres insertion, indice_par_type<tl_test_post, float>::value == 6");
   }
   void tester_length() {
      using tl_vide = type_list<>;
      using tl_test_1 = type_list<void>;
      using tl_test = type_list<int, short, int, int, char, float>;
      static_assert(length_<tl_vide>::value == 0, "length_<tl_vide>::value == 0");
      static_assert(length_<tl_test_1>::value == 1, "length_<tl_test_1>::value == 1");
      static_assert(length_<tl_test>::value == 6, "length_<tl_test>::value == 6");
      static_assert(length_<tl_test>::value + 1 == length_<static_inserer_fin<char, tl_test>::type>::value,
         "length_<tl_test>::value + 1 == length_<static_inserer_fin<char, tl_test>::type>::value");
   }
   void tester_remove_first() {
      using tl_test_1 = type_list<void>;
      using tl_test = type_list<int, short, int, int, char, float>;
      static_assert(length_<tl_test_1>::value == length_<static_remove_first_t<float, tl_test_1>>::value,
         "length_<tl_test_1>::value == length_<static_remove_first_t<float, tl_test_1>>::value");
      static_assert(length_<tl_test>::value == length_<static_remove_first_t<int, tl_test>>::value + 1,
         "length_<tl_test>::value == length_<static_remove_first_t<int, tl_test>>::value + 1");
   }
   void tester_remove() {
      using tl_test_1 = type_list<void>;
      using tl_test = type_list<int, short, int, int, char, float>;
      static_assert(length_<tl_test_1>::value == length_<static_remove_t<float, tl_test_1>>::value,
         "length_<tl_test_1>::value == length_<static_remove_t<float, tl_test_1>>::value");
      static_assert(length_<tl_test>::value == length_<static_remove_t<int, tl_test>>::value + 3,
         "length_<tl_test>::value != length_<static_remove_t<int, tl_test>>::value + 3");
   }
   void tester_type_par_indice() {
      using tl_test_1 = type_list<void>;
      using tl_test = type_list<int, short, int, int, char, float>;
      static_assert(is_same<type_par_indice<0, tl_test_1>::type, void>::value, "is_same<type_par_indice<tl_test_1, 0>::type, void>::value");
      static_assert(is_same<type_par_indice<0, tl_test>::type, int>::value, "is_same<type_par_indice<tl_test, 0>::type, int>::value");
      static_assert(is_same<type_par_indice<5, tl_test>::type, float>::value, "is_same<type_par_indice<tl_test, 5>::type, float>::value");
   }
   void tester_transform() {
      using tl_test = type_list<int, char, short>;
      using tl_res = static_transform_t<std::make_unsigned, tl_test>;
      static_assert(is_unsigned<type_par_indice<0, tl_res>::type>::value, "is_unsigned<type_par_indice<0, tl_res>::type>::value");
      static_assert(is_unsigned<type_par_indice<1, tl_res>::type>::value, "is_unsigned<type_par_indice<1, tl_res>::type>::value");
      static_assert(is_unsigned<type_par_indice<2, tl_res>::type>::value, "is_unsigned<type_par_indice<2, tl_res>::type>::value");
   }
}

void tester_algorithmes_statiques() {
   ::tester_count_consecutive();
   ::tester_count_if();
   ::tester_indice_par_type();
   ::tester_inserer();
   ::tester_inserer_fin();
   ::tester_length();
   ::tester_remove_first();
   ::tester_remove();
   ::tester_type_par_indice();
   ::tester_transform();
}
