#ifndef STATIC_ALGORITHMS_H
#define STATIC_ALGORITHMS_H

#include "types_brefs.h"
#include "type_list.h"
#include <type_traits>

//
// -=-=-=-=-=-=-=-=-
// Static algorithms without dependencies
// -=-=-=-=-=-=-=-=-
//
template <class TList>
   struct length_;
template <class ... Ts>
   struct length_<type_list<Ts...>> : int_<sizeof...(Ts)> {
   };

//
// Fusionner deux listes de types
//
template <class TL0, class TL1>
   struct static_merge;
template <class ... TL0, class ... TL1>
   struct static_merge<type_list<TL0...>, type_list<TL1...>> {
      using type = type_list<TL0..., TL1 ...>;
   };
template <class T, class ... TL>
   struct static_merge<T, type_list<TL...>> {
      using type = type_list<T, TL ...>;
   };
template <class T, class ... TL>
   struct static_merge<type_list<TL...>, T> {
      using type = type_list<TL ..., T>;
   };

template <class TL0, class TL1>
   using static_merge_t = typename static_merge<TL0, TL1>::type;

//
// count_consecutive_beginning<T, TList>::value est le nombre d'occurrences
// conscutives de T au dbut de TList
//
template <class T, class TL>
struct count_consecutive_beginning;
template <class T, class ... Q>
struct count_consecutive_beginning<T, type_list<T, Q...>>
   : int_<1 + count_consecutive_beginning<T, type_list<Q...>>::value> {
};
template <class U, class ...T>
   struct count_consecutive_beginning<U, type_list<T...>> : int_<0> {
   };
template <class T>
   struct count_consecutive_beginning<T, type_list<T>> : int_<1> {
   };
template <class U, class T>
   struct count_consecutive_beginning<U, type_list<T>> : int_<0> {
   };

//
// Trouver l'indice de la 1re occurrence d'un type dans une liste de types
//
template <class T, class TList>
   struct indice_par_type : int_<-1> {
   };
template <class T, class ...Q>
   struct indice_par_type<T, type_list<T, Q...>> : int_<0> {
   };
template <class U, class T, class ...Q>
   struct indice_par_type<U, type_list<T, Q...>>
      : std::conditional_t<indice_par_type<U, type_list<Q...>>::value == -1,
                           int_<-1>,
                           int_<1 + indice_par_type<U, type_list<Q...>>::value>> {
   };

//
// Insre un type  la fin d'une liste de types
//
// Depends on: nil
//
template <class T, class TL>
   struct static_inserer_fin;
template <class U, class T, class ... Q>
   struct static_inserer_fin<U, type_list<T, Q...>> {
      using type = type_list <T, Q..., U>;
   };
template <class T, class U>
   struct static_inserer_fin<U, type_list<T>> {
      using type = type_list <T, U>;
   };

//
// Compte le nombre d'lments d'une liste pour lesquels un
// prdicat s'avre (important: on considre 1 comme vrai et
// 0 comme faux)
//
// Depends on: nil
//
template <template <class> class Pred, class TList>
   struct static_count_if;
template <template <class> class Pred, class T, class ... Q>
   struct static_count_if<Pred, type_list<T, Q...>>
      : int_<static_cast<int>(Pred<T>::value) + static_count_if<Pred, type_list<Q...>>::value> {
   };
template <template <class> class Pred, class T>
   struct static_count_if<Pred, type_list<T>> : int_<static_cast<int>(Pred<T>::value)> {
   };

//
// Supprimer la premiere occurrence d'un type d'une liste de types
//
template <class T, class TL>
   struct static_remove_first;

template <class T, class TL>
   using static_remove_first_t = typename static_remove_first<T, TL>::type;

template <class U, class T, class ... Q>
   struct static_remove_first<U, type_list<T, Q...>> {
      using type = static_merge_t<T, static_remove_first_t<U, type_list<Q...>>>;
   };
template <class T, class ... Q>
   struct static_remove_first<T, type_list<T, Q...>> {
      using type = type_list<Q...>;
   };
template <class U, class T>
   struct static_remove_first<U, type_list<T>> {
      using type = type_list<T>;
   };
template <class T>
   struct static_remove_first<T, type_list<>> {
      using type = type_list<>;
   };

//
// static_remove_if<Pred, TL>::type est une liste de types dont toutes les occurrences
// de types T tels que Pred<T>::value s'avre ont t supprimes
//
template <template <class> class Pred, class TList>
   struct static_remove_if;

template <template <class> class Pred, class TList>
   using static_remove_if_t = typename static_remove_if<Pred, TList>::type;

template <template <class> class Pred, class T, class ... Q>
   struct static_remove_if<Pred, type_list<T, Q...>> {
      using type = std::conditional_t<
         Pred<T>::value,
         static_remove_if_t<Pred, type_list<Q...>>,
         static_merge_t<T, static_remove_if_t<Pred, type_list<Q...>>>
      >;
   };
template <template <class> class Pred, class T>
   struct static_remove_if<Pred, type_list<T>> {
      using type = std::conditional_t<Pred<T>::value, type_list<>, type_list<T>>;
   };


//
// Prdicat statique is_<T>
// - is_<T>::type est un prdicat sur un type U qui s'avre si T est U
//
template <class T>
struct is_ {
   template <class U>
      struct type : bool_<std::is_same<T, U>::value> {
      };
};
//
// static_remove<T,TL> est une liste de types telle que TL dont on aurait
// supprim toutes les occurrences de T
//
// elle dpend de static_remove_if et de is_
//
template <class T, class TList>
   struct static_remove {
      // template <class U> using test = typename is_<T>::type;
      using type = static_remove_if_t<::is_<T>::template type, TList>; // ::is_ due to a known Clang bug
   };

template <class T, class TList>
   using static_remove_t = typename static_remove<T, TList>::type;

//
// Trouver un type par son indice dans une liste de types
//
// Depends on: nil
//
template <unsigned int, class TList>
   struct type_par_indice;

template <unsigned int i, class TL>
   using type_par_indice_t = typename type_par_indice<i, TL>::type;

template <class T, class ... Q>
   struct type_par_indice<0, type_list<T, Q...>> {
      using type = T;
   };
template <unsigned int i, class T, class ... Q>
   struct type_par_indice<i, type_list<T, Q...>> {
      static_assert(i <= sizeof...(Q), "i <= sizeof...(Q)");
      using type = type_par_indice_t<i - 1, type_list<Q...>>;
   };

//
// begins_with_n<N, T, TList>::value est vrai seulement si
// on trouve au moins N occurrences de T au dbut de TList
//
template <int N, class T, class Tlist>
   struct begins_with_n;
template <int N, class T, class ... Q>
   struct begins_with_n<N, T, type_list<T, Q...>>
      : bool_<begins_with_n<N - 1, T, type_list<Q...>>::value> {
   };
template <class T, class ... Q>
   struct begins_with_n<1, T, type_list<T, Q...>> : std::true_type
   {
   };
template <int N, class U, class T, class ... Q>
   struct begins_with_n<N, U, type_list<T, Q...>> : std::false_type {
   };
template <int N, class U>
   struct begins_with_n<N, U, type_list<>> : std::false_type {
   };

//
// static_transform<F, TList>
// - applique F<T>  chaque T de la liste TList
// - le rsultat de F<T> est un type
// - le rsultat de static_transform<F, TList> est static_transform<F, TList>::type,
//   une liste de types correspondant lment par lment aux types dans TList, une
//   fois ceux-ci transforms par F
//
template <template <class> class F, class TList>
   struct static_transform;

template <template <class> class F, class TL>
   using static_transform_t = typename static_transform<F, TL>::type;

template <template <class> class F, class T, class ... Q>
   struct static_transform<F, type_list<T, Q...>> {
      using type = static_merge_t<typename F<T>::type, static_transform_t<F, type_list<Q...>>>;
   };
template <template <class> class F, class T>
   struct static_transform<F, type_list<T>> {
      using type = type_list<typename F<T>::type>;
   };

template <template <class, class> class Op, class Init, class TList>
   struct static_accumulate;
template <template <class, class> class Op, class Init, class T, class ... Q>
   struct static_accumulate<Op, Init, type_list<T, Q...>> {
      using type = typename Op<T, typename static_accumulate<Op, Init, type_list<Q...>>::type>::type;
   };
template <template <class, class> class Op, class Init, class T>
   struct static_accumulate<Op, Init, type_list<T>> {
      using type = typename Op<T, Init>::type;
   };

template <template <class, class> class Op, class Init, class TList>
   using static_accumulate_t = typename static_accumulate<Op, Init, TList>::type;

//
// static_for_each<TList>::appliquer(Op)
//
// Applique l'opration Op<T>()  chaque type T de la liste TList
//
// Depends on: nil
//
template <class TL>
   struct static_for_each;
template <class T, class ... Ts>
   struct static_for_each<type_list<T, Ts...>> {
      template <class F>
         static void appliquer(F f) {
            f(T{});
            static_for_each<type_list<Ts...>>::appliquer(f);
         }
   };
template <class T>
   struct static_for_each<type_list<T>> {
      template <class F>
         static void appliquer(F f) {
            f(T{});
         }
   };

//
// behead<T, Tlist>::type est la liste TList de laquelle on a
// lagu toutes les occurrences de T au dbut
//
template <class T, class TList>
   struct behead;

template <class T, class TL>
   using behead_t = typename behead<T, TL>::type;

template <class T, class ... Q>
   struct behead<T, type_list<T, Q...>> {
      using type = behead_t<T, type_list<Q...>>;
   };
template <class U, class T, class ... Q>
   struct behead<U, type_list<T, Q...>> {
      using type = type_list<T, Q...>;
   };
template <class T>
   struct behead<T, type_list<>> {
      using type = type_list<>;
   };

//
// -=-=-=-=-=-=-=-=-
// Static algorithms with little dependencies
// -=-=-=-=-=-=-=-=-
//

#include "outils_numeriques_statiques.h"

//
// static_sum<TList>
// - reprsente la somme des lments de la liste TList
// - le fruit de static_sum<TList> est static_sum<TList>::type, un integer_constant
//   reprsentant la valeur de la somme calcule
//
// Depends on:
// - static_accumulate
// - statc_add
//
template <class TList>
   struct static_sum : static_accumulate_t<static_add, static_add_base::neuter, TList> {
   };

//
// make_int_list<Debut, Fin>::type est la liste des types
// int_<Debut>, int_<Debut+1>, ..., int_<Fin> inclusivement
//
// ICI : migrer vers std::make_integer_sequence

template <int Debut, int Fin>
   struct make_int_list {
      static_assert(Debut < Fin, "Debut < Fin");
      using type = static_merge_t<
         int_<Debut>, typename make_int_list<Debut + 1, Fin>::type
      >;
   };
template <int Fin>
   struct make_int_list<Fin, Fin> {
      using type = int_<Fin>;
   };



//
// est_dans<T,TList>::value est vrai ssi T est dans TList
//
// Depends on: indice_par_type
//
template <class T, class ... TList>
   struct est_dans : bool_<indice_par_type<T, type_list<TList...>>::value != -1> {
   };

//
// Pivoter les lments d'une liste de types vers la gauche, de
// manire cyclique (en dplaant le 1er lment de la liste  la
// fin)
//
// Depends on: inserer_fin
//
template <class TList>
   struct static_pivoter_gauche;
template <class T, class ... Q>
   struct static_pivoter_gauche<type_list<T, Q...>> {
      using type = typename static_inserer_fin<T, type_list<Q...>>::type;
   };
template <class T>
   struct static_pivoter_gauche<type_list<T>> {
      using type = type_list<T>;
   };

//
// -=-=-=-=-=-=-=-=-
// Static algorithms with strong dependencies
// -=-=-=-=-=-=-=-=-
//

//
// make_list<N, T>::type est une liste comprenant N occurrences
// du type T
//
template <int N, class T>
   struct make_list {
      static_assert(N > 0, "invalid list length");
      using type = static_merge_t<T, typename make_list<N - 1, T>::type>;
   };
template <class T>
   struct make_list<1, T> {
      using type = type_list<T>;
   };
template <int N, class T>
   using make_list_t = typename make_list<N, T>::type;

//
// static_first_n<N, TList>::type est la liste des N premiers lments de TList
// static_first_if<Pred, TList>::type est la liste des lments T en tte de TList tels que Pred<T>::value
//
template <int N, class TList>
   struct static_first_n;

template <int N, class TList>
   using static_first_n_t = typename static_first_n<N, TList>::type;

template <int N, class T, class ... Q>
   struct static_first_n<N, type_list<T, Q...>> {
      static_assert(N <= length_<type_list<T, Q...>>::value, "list too short");
      using type = static_merge_t<T, static_first_n_t<N - 1, type_list<Q...>>>;
   };
template <class T, class ... Q>
   struct static_first_n<1, type_list<T, Q...>> {
      using type = type_list<T>;
   };
template <class T, class ... Q>
   struct static_first_n<0, type_list<T, Q...>> {
      using type = type_list<>;
   };


template <template <class> class Pred, class TList>
   struct static_first_if;
template <template <class> class Pred, class TList>
   using static_first_if_t = typename static_first_if<Pred, TList>::type;
template <template <class> class Pred, class T, class ... Ts>
   struct static_first_if<Pred, type_list<T, Ts...>> {
      using type = std::conditional_t <
         Pred<T>::value,
         static_merge_t<T, static_first_if_t<Pred, type_list<Ts...>>>,
         type_list<>
      >;
   };
template <template <class> class Pred>
   struct static_first_if<Pred, type_list<>> {
      using type = type_list<>;
   };


//
// static_remove_tail<TList,N>::type est la liste TList de laquelle
// ont t amputs les N derniers lments
//
// Depends on:
// - static_first_n
// - length_
//
template <int N, class TList>
   struct static_remove_tail;
template <int N, class T, class ... Q>
   struct static_remove_tail<N, type_list<T, Q...>> {
      using type = static_first_n_t<
         static_max<int_<length_<type_list<T, Q...>>::value - N>, int_<0>>::value,
         type_list<T, Q...>
      >;
   };


//
// behead_n<N, TList>::type quivaut  la liste TList
// de laquelle on aurait amput les N premiers lments
//
// Depends on:
// - length_
// - static_tail
//
template <int N, class TList>
   struct behead_n;

template <int N, class TL>
   using behead_n_t = typename behead_n<N, TL>::type;

template <int N, class ... T>
   struct behead_n<N, type_list<T...>> {
      static_assert(N >= 0 && length_<type_list<T...>>::value >= N, "list too short");
      using type = behead_n_t<
         N - 1, typename static_tail<type_list<T...>>::type
      >;
   };
template <class ... T>
   struct behead_n<0, type_list<T...>> {
      using type = type_list<T...>;
   };
template <class T>
   struct behead_n<0, T> {
      using type = type_list<T>;
   };

//
// static_sublist<B,E,TList>::type is the list made from
// the sub list (B..E( in TList
//
// Depends on:
// - behead_n
// - length_
// - static_remove_tail
//
template <int B, int E, class TList>
   class static_sublist {
      static_assert(B < E, "bornes invalides");
      static_assert(length_<TList>::value >= E - B, "list too short");
      using beheaded_list = behead_n_t<B, TList>;
   public:
      using type = static_first_n_t<E - B, beheaded_list>;
   };

template <int B, int E, class TL>
   using static_sublist_t = typename static_sublist<B, E, TL>::type;

//
//
//
template <class T, class U>
   struct static_concatenate {
      using type = type_list<T, U>;
   };
template <class T, class ... Ts>
   struct static_concatenate<T, type_list<Ts...>> {
      using type = type_list<T, Ts...>;
   };
template <class T>
   struct static_concatenate<T, type_list<>> {
      using type = type_list<T>;
   };
template <class T, class U>
   using static_concatenate_t = typename static_concatenate<T, U>::type;

template <int N, class TL>
   struct static_split_equal_parts;
template <int N, class TL>
   using static_split_equal_parts_t = typename static_split_equal_parts<N, TL>::type;
template <int N, class T, class ... Ts>
   struct static_split_equal_parts<N, type_list<T, Ts...>> {
      static_assert(length_<type_list<T, Ts...>>::value % N == 0, "Liste de taille incorrecte");
      using type = static_concatenate_t <
         static_first_n_t<N, type_list<T, Ts...>>,
         static_split_equal_parts_t<
         N, behead_n_t<N, type_list<T, Ts...>>
         >
      >;
   };
template <int N>
   struct static_split_equal_parts<N, type_list<>> {
      using type = type_list<>;
   };

//
//
//

template <int Pos, template <class> class Pred, class TL>
   struct static_find_indices_in_list_if_impl;
template <int Pos, template <class> class Pred, class T, class ... Ts>
   struct static_find_indices_in_list_if_impl<Pos, Pred, type_list<T, Ts...>> {
      using type = std::conditional_t<
         Pred<type_list<T, Ts...>>::value,
         typename static_merge<int_<Pos>, typename static_find_indices_in_list_if_impl<Pos + 1, Pred, type_list<Ts...>>::type>::type,
         typename static_find_indices_in_list_if_impl<Pos + 1, Pred, type_list<Ts...>>::type
      >;
   };
template <int Pos, template <class> class Pred>
   struct static_find_indices_in_list_if_impl<Pos, Pred, type_list<>> {
      using type = type_list<>;
   };

template <template <class> class Pred, class TL>
   struct static_find_indices_in_list_if : static_find_indices_in_list_if_impl<0, Pred, TL> {
   };

//
//
//
void tester_algorithmes_statiques();

#endif
